/*
DingusPPC - The Experimental PowerPC Macintosh emulator
Copyright (C) 2018-24 divingkatae and maximum
                      (theweirdo)     spatium

(Contact divingkatae#1017 or powermax#2286 on Discord for more info)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include <core/timermanager.h>
#include <loguru.hpp>
#include "ppcemu.h"
#include "ppcmmu.h"
#include "ppcdisasm.h"
#include "ppcoptablemacros.h"

#include <algorithm>
#include <cstring>
#include <iostream>
#include <map>
#include <setjmp.h>
#include <stdexcept>
#include <stdio.h>
#include <string>

#ifdef __APPLE__
#include <mach/mach_time.h>
#undef EXC_SYSCALL
static struct mach_timebase_info timebase_info;
static uint64_t
ConvertHostTimeToNanos2(uint64_t host_time)
{
    if (timebase_info.numer == timebase_info.denom)
        return host_time;
    long double answer = host_time;
    answer *= timebase_info.numer;
    answer /= timebase_info.denom;
    return (uint64_t)answer;
}
#endif

using namespace std;
using namespace dppc_interpreter;

MemCtrlBase* mem_ctrl_instance = 0;

bool is_601 = false;

bool power_on = false;
Po_Cause power_off_reason = po_enter_debugger;

SetPRS ppc_state;

uint32_t ppc_next_instruction_address;    // Used for branching, setting up the NIA

unsigned exec_flags; // execution control flags
// FIXME: exec_timer is read by main thread ppc_main_opcode;
// written by audio dbdma DMAChannel::update_irq .. add_immediate_timer
volatile bool exec_timer;
bool int_pin = false; // interrupt request pin state: true - asserted
bool dec_exception_pending = false;

/* copy of local variable bb_start_la. Need for correct
   calculation of CPU cycles after setjmp that clobbers
   non-volatile local variables. */
uint32_t    glob_bb_start_la;

/* variables related to virtual time */
const bool g_realtime = false;
uint64_t g_nanoseconds_base;
uint64_t g_icycles;
int      icnt_factor;

/* global variables related to the timebase facility */
uint64_t tbr_wr_timestamp;  // stores vCPU virtual time of the last TBR write
uint64_t rtc_timestamp;     // stores vCPU virtual time of the last RTC write
uint64_t tbr_wr_value;      // last value written to the TBR
uint32_t tbr_freq_ghz;      // TBR/RTC driving frequency in GHz expressed as a
                            // 32 bit fraction less than 1.0 (999.999999 MHz maximum).
uint64_t tbr_period_ns;     // TBR/RTC period in ns expressed as a 64 bit value
                            // with 32 fractional bits (<1 Hz minimum).
uint64_t timebase_counter;  // internal timebase counter
uint64_t dec_wr_timestamp;  // stores vCPU virtual time of the last DEC write
uint32_t dec_wr_value;      // last value written to the DEC register
uint32_t rtc_lo;            // MPC601 RTC lower, counts nanoseconds
uint32_t rtc_hi;            // MPC601 RTC upper, counts seconds

#ifdef CPU_PROFILING

/* global variables for lightweight CPU profiling */
uint64_t num_executed_instrs;
uint64_t num_supervisor_instrs;
uint64_t num_int_loads;
uint64_t num_int_stores;
uint64_t exceptions_processed;
#ifdef CPU_PROFILING_OPS
std::unordered_map<uint32_t, uint64_t> num_opcodes;
#endif

#include "utils/profiler.h"
#include <memory>

class CPUProfile : public BaseProfile {
public:
    CPUProfile() : BaseProfile("PPC_CPU") {};

    void populate_variables(std::vector<ProfileVar>& vars) {
        vars.clear();

        vars.push_back({.name = "Executed Instructions Total",
                        .format = ProfileVarFmt::DEC,
                        .value = num_executed_instrs});

        vars.push_back({.name = "Executed Supervisor Instructions",
                        .format = ProfileVarFmt::DEC,
                        .value = num_supervisor_instrs});

        vars.push_back({.name = "Integer Load Instructions",
                        .format = ProfileVarFmt::DEC,
                        .value = num_int_loads});

        vars.push_back({.name = "Integer Store Instructions",
                        .format = ProfileVarFmt::DEC,
                        .value = num_int_stores});

        vars.push_back({.name = "Exceptions processed",
                        .format = ProfileVarFmt::DEC,
                        .value = exceptions_processed});

        // Generate top N op counts with readable names.
#ifdef CPU_PROFILING_OPS
        PPCDisasmContext ctx;
        ctx.instr_addr = 0;
        ctx.simplified = false;
        std::vector<std::pair<std::string, uint64_t>> op_name_counts;
        for (const auto& pair : num_opcodes) {
            ctx.instr_code = pair.first;
            auto op_name = disassemble_single(&ctx);
            op_name_counts.emplace_back(op_name, pair.second);
        }
        size_t top_ops_size = std::min(op_name_counts.size(), size_t(20));
        std::partial_sort(op_name_counts.begin(), op_name_counts.begin() + top_ops_size, op_name_counts.end(), [](const auto& a, const auto& b) {
            return b.second < a.second;
        });
        op_name_counts.resize(top_ops_size);
        for (const auto& pair : op_name_counts) {
            vars.push_back({.name = "Instruction " + pair.first,
                            .format = ProfileVarFmt::COUNT,
                            .value = pair.second,
                            .count_total = num_executed_instrs});
        }
#endif
    };

    void reset() {
        num_executed_instrs = 0;
        num_supervisor_instrs = 0;
        num_int_loads = 0;
        num_int_stores = 0;
        exceptions_processed = 0;
#ifdef CPU_PROFILING_OPS
        num_opcodes.clear();
#endif
    };
};

#endif

/** Opcode lookup tables. */

/** Primary opcode (bits 0...5) lookup table. */
static PPCOpcode OpcodeGrabber[64];

/** Lookup tables for branch instructions. */
static PPCOpcode Opcode16Grabber[4];    // bcla

static PPCOpcode Opcode18Grabber[4];     // bla

/** Instructions decoding tables for integer,
    single floating-point, and double-floating point ops respectively */

static PPCOpcode Opcode19Grabber[2048];
static PPCOpcode Opcode31Grabber[2048];
static PPCOpcode Opcode59Grabber[64];
static PPCOpcode Opcode63Grabber[2048];

/** Exception helpers. */

void ppc_illegalop(uint32_t null_val) {
    ppc_exception_handler(Except_Type::EXC_PROGRAM, Exc_Cause::ILLEGAL_OP);
}

void ppc_fpu_off() {
    ppc_exception_handler(Except_Type::EXC_PROGRAM, Exc_Cause::FPU_OFF);
}

void ppc_assert_int() {
    int_pin = true;
    if (ppc_state.msr & MSR::EE) {
        LOG_F(5, "CPU ExtIntHandler called");
        ppc_exception_handler(Except_Type::EXC_EXT_INT, 0);
    } else {
        LOG_F(5, "CPU IRQ ignored!");
    }
}

void ppc_release_int() {
    int_pin = false;
}

/** Opcode decoding functions. */

static void ppc_opcode16(uint32_t subop, uint32_t instr) {
    Opcode16Grabber[instr & 3](instr);
}

static void ppc_opcode18(uint32_t instr) {
    Opcode18Grabber[instr & 3](instr);
}

template<field_601 for601>
void ppc_opcode19(uint32_t instr) {
    uint16_t subop_grab = instr & 0x7FF;

    switch (subop_grab) {
    case 0:
        ppc_mcrf(instr);
        break;
    case 32:
        ppc_bclr<LK0>(instr);
        break;
    case 33:
        ppc_bclr<LK1>(instr);
        break;
    case 66:
        ppc_crnor(instr);
        break;
    case 100:
        ppc_rfi(instr);
        break;
    case 258:
        ppc_crandc(instr);
        break;
    case 300:
        ppc_isync(instr);
        break;
    case 386:
        ppc_crxor(instr);
        break;
    case 450:
        ppc_crnand(instr);
        break;
    case 514:
        ppc_crand(instr);
        break;
    case 578:
        ppc_creqv(instr);
        break;
    case 834:
        ppc_crorc(instr);
        break;
    case 898:
        ppc_cror(instr);
        break;
    case 1056:
        ppc_bcctr<LK0, for601>(instr);
        break;
    case 1057:
        ppc_bcctr<LK1, for601>(instr);
        break;
    default:
        ppc_illegalop();
    }
}

template void ppc_opcode19<NOT601>(uint32_t);
template void ppc_opcode19<IS601>(uint32_t);

void ppc_opcode31(uint32_t instr) {
}

void ppc_opcode59(uint32_t instr) {
    uint16_t subop_grab = instr & 0x3FUL;
    Opcode59Grabber[subop_grab](instr);
}

void ppc_opcode63(uint32_t instr) {
    uint16_t subop_grab = instr & 0x7FFUL;
    Opcode63Grabber[subop_grab](instr);
}

/* Dispatch using main opcode */
void ppc_main_opcode(uint32_t instr) {
#ifdef CPU_PROFILING
    num_executed_instrs++;
#if defined(CPU_PROFILING_OPS)
    num_opcodes[instr]++;
#endif
#endif
    uint32_t opcode = (instr >> 26) & 63;
    switch (opcode) {
        case 16: 
            Opcode16Grabber[instr & 3](instr);
            break;
        case 18:
            Opcode16Grabber[instr & 3](instr); 
            break;
        case 19:
            if (is_601)
                ppc_opcode19<IS601>(instr);
            else
                ppc_opcode19<NOT601>(instr);
            break;
        case 31:
            uint16_t subop_grab = instr & 0x7FFUL;
            Opcode31Grabber[subop_grab](instr);
            break;
        case 59:
            uint16_t subop_grab = instr & 0x7FFUL;
            Opcode31Grabber[subop_grab](instr);
            break;
        case 63:
            uint16_t subop_grab = instr & 0x7FFUL;
            Opcode31Grabber[subop_grab](instr);
            break;
        default:
            OpcodeGrabber[opcode](instr);
    }

#ifdef CPU_PROFILING
    uint32_t load_chk = (instr >> 26);
    if ((load_chk > 31) && (load_chk < 59))
        num_int_loads++;
#endif
}

static long long cpu_now_ns() {
#ifdef __APPLE__
    return ConvertHostTimeToNanos2(mach_absolute_time());
#else
    return std::chrono::duration_cast<std::chrono::nanoseconds>(
        std::chrono::high_resolution_clock::now().time_since_epoch()).count();
#endif
}

uint64_t get_virt_time_ns()
{
    if (g_realtime) {
        return cpu_now_ns() - g_nanoseconds_base;
    } else {
        return g_icycles << icnt_factor;
    }
}

static uint64_t process_events()
{
    exec_timer = false;
    uint64_t slice_ns = TimerManager::get_instance()->process_timers();
    if (slice_ns == 0) {
        // execute 25.000 cycles
        // if there are no pending timers
        return g_icycles + 25000;
    }
    return g_icycles + (slice_ns >> icnt_factor) + 1;
}

static void force_cycle_counter_reload()
{
    // tell the interpreter loop to reload cycle counter
    exec_timer = true;
}

/** Execute PPC code as long as power is on. */
// inner interpreter loop
static void ppc_exec_inner()
{
    uint64_t max_cycles;
    uint32_t page_start, eb_start, eb_end;
    uint8_t* pc_real;
    uint32_t instr = 0;
    uint32_t i_params[5] = {0};
    double f_params[4] = {0};

    max_cycles = 0;

    while (power_on) {
        // define boundaries of the next execution block
        // max execution block length = one memory page
        eb_start   = ppc_state.pc;
        page_start = eb_start & PPC_PAGE_MASK;
        eb_end     = page_start + PPC_PAGE_SIZE - 1;
        exec_flags = 0;

        pc_real = mmu_translate_imem(eb_start);
        instr   = ppc_set_cur_instruction(pc_real);

        // interpret execution block
        while (power_on && ppc_state.pc < eb_end) {
            ppc_main_opcode(instr);
            if (g_icycles++ >= max_cycles || exec_timer) {
                max_cycles = process_events();
            }

            if (exec_flags) {
                // define next execution block
                eb_start = ppc_next_instruction_address;
                if (!(exec_flags & EXEF_RFI) && (eb_start & PPC_PAGE_MASK) == page_start) {
                    pc_real += (int)eb_start - (int)ppc_state.pc;
                    instr    = ppc_set_cur_instruction(pc_real);
                } else {
                    page_start = eb_start & PPC_PAGE_MASK;
                    eb_end = page_start + PPC_PAGE_SIZE - 1;
                    pc_real    = mmu_translate_imem(eb_start);
                    instr      = ppc_set_cur_instruction(pc_real);
                }
                ppc_state.pc = eb_start;
                exec_flags = 0;
            } else {
                ppc_state.pc += 4;
                pc_real      += 4;
                instr         = ppc_set_cur_instruction(pc_real);
            }
        }
    }
}

// outer interpreter loop
void ppc_exec()
{
    if (setjmp(exc_env)) {
        // process low-level exceptions
        //LOG_F(9, "PPC-EXEC: low_level exception raised!");
        ppc_state.pc = ppc_next_instruction_address;
    }

    while (power_on) {
        ppc_exec_inner();
    }
}

/** Execute one PPC instruction. */
void ppc_exec_single()
{
    if (setjmp(exc_env)) {
        // process low-level exceptions
        //LOG_F(9, "PPC-EXEC: low_level exception raised!");
        ppc_state.pc = ppc_next_instruction_address;
        exec_flags = 0;
        return;
    }

    mmu_translate_imem(ppc_state.pc);

    uint8_t* pc_real = mmu_translate_imem(ppc_state.pc);
    uint32_t instr   = ppc_set_cur_instruction(pc_real);
    ppc_main_opcode(instr);

    g_icycles++;
    process_events();

    if (exec_flags) {
        ppc_state.pc = ppc_next_instruction_address;
        exec_flags = 0;
    } else {
        ppc_state.pc += 4;
    }
}

/** Execute PPC code until goal_addr is reached. */

// inner interpreter loop
static void ppc_exec_until_inner(const uint32_t goal_addr)
{
    uint64_t max_cycles;
    uint32_t page_start, eb_start, eb_end;
    uint8_t* pc_real;
    uint32_t instr;

    max_cycles = 0;

    do {
        // define boundaries of the next execution block
        // max execution block length = one memory page
        eb_start   = ppc_state.pc;
        page_start = eb_start & PPC_PAGE_MASK;
        eb_end     = page_start + PPC_PAGE_SIZE - 1;
        exec_flags = 0;

        pc_real = mmu_translate_imem(eb_start);
        instr   = ppc_set_cur_instruction(pc_real);

        // interpret execution block
        while (power_on && ppc_state.pc < eb_end) {
            ppc_main_opcode(instr);
            if (g_icycles++ >= max_cycles || exec_timer) {
                max_cycles = process_events();
            }

            if (exec_flags) {
                // define next execution block
                eb_start = ppc_next_instruction_address;
                if (!(exec_flags & EXEF_RFI) && (eb_start & PPC_PAGE_MASK) == page_start) {
                    pc_real += (int)eb_start - (int)ppc_state.pc;
                    instr    = ppc_set_cur_instruction(pc_real);
                } else {
                    page_start = eb_start & PPC_PAGE_MASK;
                    eb_end = page_start + PPC_PAGE_SIZE - 1;
                    pc_real    = mmu_translate_imem(eb_start);
                    instr      = ppc_set_cur_instruction(pc_real);
                }
                ppc_state.pc = eb_start;
                exec_flags = 0;
            } else {
                ppc_state.pc += 4;
                pc_real += 4;
                instr = ppc_set_cur_instruction(pc_real);
            }

            if (ppc_state.pc == goal_addr)
                break;
        }
    } while (power_on && ppc_state.pc != goal_addr);
}

// outer interpreter loop
void ppc_exec_until(volatile uint32_t goal_addr)
{
    if (setjmp(exc_env)) {
        // process low-level exceptions
        //LOG_F(9, "PPC-EXEC: low_level exception raised!");
        ppc_state.pc = ppc_next_instruction_address;
    }

    do {
        ppc_exec_until_inner(goal_addr);
    } while (power_on && ppc_state.pc != goal_addr);
}

/** Execute PPC code until control is reached the specified region. */

// inner interpreter loop
static void ppc_exec_dbg_inner(const uint32_t start_addr, const uint32_t size)
{
    uint64_t max_cycles;
    uint32_t page_start, eb_start, eb_end;
    uint8_t* pc_real;
    uint32_t instr = 0;

    max_cycles = 0;

    while (power_on && (ppc_state.pc < start_addr || ppc_state.pc >= start_addr + size)) {
        // define boundaries of the next execution block
        // max execution block length = one memory page
        eb_start   = ppc_state.pc;
        page_start = eb_start & PPC_PAGE_MASK;
        eb_end     = page_start + PPC_PAGE_SIZE - 1;
        exec_flags = 0;

        pc_real = mmu_translate_imem(eb_start);
        instr   = ppc_set_cur_instruction(pc_real);

        // interpret execution block
        while (power_on && (ppc_state.pc < start_addr || ppc_state.pc >= start_addr + size)
                && (ppc_state.pc < eb_end)) {
            ppc_main_opcode(instr);
            if (g_icycles++ >= max_cycles || exec_timer) {
                max_cycles = process_events();
            }

            if (exec_flags) {
                // define next execution block
                eb_start = ppc_next_instruction_address;
                if (!(exec_flags & EXEF_RFI) && (eb_start & PPC_PAGE_MASK) == page_start) {
                    pc_real += (int)eb_start - (int)ppc_state.pc;
                    instr    = ppc_set_cur_instruction(pc_real);
                } else {
                    page_start = eb_start & PPC_PAGE_MASK;
                    eb_end = page_start + PPC_PAGE_SIZE - 1;
                    pc_real    = mmu_translate_imem(eb_start);
                    instr      = ppc_set_cur_instruction(pc_real);
                }
                ppc_state.pc = eb_start;
                exec_flags = 0;
            } else {
                ppc_state.pc += 4;
                pc_real += 4;
                instr    = ppc_set_cur_instruction(pc_real);
            }
        }
    }
}

// outer interpreter loop
void ppc_exec_dbg(volatile uint32_t start_addr, volatile uint32_t size)
{
    if (setjmp(exc_env)) {
        // process low-level exceptions
        //LOG_F(9, "PPC-EXEC: low_level exception raised!");
        ppc_state.pc = ppc_next_instruction_address;
    }

    while (power_on && (ppc_state.pc < start_addr || ppc_state.pc >= start_addr + size)) {
        ppc_exec_dbg_inner(start_addr, size);
    }
}

/*
Opcode table macros:
- d is for dot (RC).
- o is for overflow (OV).
- c is for carry CARRY0/CARRY1. It also works for other options:
  SHFT0/SHFT1, RIGHT0/LEFT1, uint8_t/uint16_t/uint32_t, and int8_t/int16_t.
 */

#define OP(opcode, fn) \
do { \
    OpcodeGrabber[opcode] = fn; \
} while (0)

#define OPX(opcode, subopcode, fn) \
do { \
    opcode ## Grabber[((subopcode)<<1)] = fn; \
} while (0)

#define OPXd(opcode, subopcode, fn) \
do { \
    opcode ## Grabber[((subopcode)<<1)] = fn<RC0>; \
    opcode ## Grabber[((subopcode)<<1)+1] = fn<RC1>; \
} while (0)

#define OPXod(opcode, subopcode, fn) \
do { \
    opcode ## Grabber[((subopcode)<<1)] = fn<RC0, OV0>; \
    opcode ## Grabber[((subopcode)<<1)+1] = fn<RC1, OV0>; \
    opcode ## Grabber[1024+((subopcode)<<1)] = fn<RC0, OV1>; \
    opcode ## Grabber[1024+((subopcode)<<1)+1] = fn<RC1, OV1>; \
} while (0)

#define OPXdc(opcode, subopcode, fn, carry) \
do { \
    opcode ## Grabber[((subopcode)<<1)] = fn<carry, RC0>; \
    opcode ## Grabber[((subopcode)<<1)+1] = fn<carry, RC1>; \
} while (0)

#define OPXcod(opcode, subopcode, fn, carry) \
do { \
    opcode ## Grabber[((subopcode)<<1)] = fn<carry, RC0, OV0>; \
    opcode ## Grabber[((subopcode)<<1)+1] = fn<carry, RC1, OV0>; \
    opcode ## Grabber[1024+((subopcode)<<1)] = fn<carry, RC0, OV1>; \
    opcode ## Grabber[1024+((subopcode)<<1)+1] = fn<carry, RC1, OV1>; \
} while (0)

#define OP31(subopcode, fn) OPX(Opcode31, subopcode, fn)
#define OP31d(subopcode, fn) OPXd(Opcode31, subopcode, fn)
#define OP31od(subopcode, fn) OPXod(Opcode31, subopcode, fn)
#define OP31dc(subopcode, fn, carry) OPXdc(Opcode31, subopcode, fn, carry)
#define OP31cod(subopcode, fn, carry) OPXcod(Opcode31, subopcode, fn, carry)

#define OP59d(subopcode, fn) OPXd(Opcode59, subopcode, fn)

#define OP63(subopcode, fn) OPX(Opcode63, subopcode, fn)
#define OP63d(subopcode, fn) OPXd(Opcode63, subopcode, fn)
#define OP63dc(subopcode, fn, carry) OPXdc(Opcode63, subopcode, fn, carry)

inline void initialize_ppc_opcode_tables(bool include_601) {
    std::fill_n(OpcodeGrabber, 64, ppc_illegalop);
    std::fill_n(Opcode16Grabber, 3, ppc_illegalop);
    std::fill_n(Opcode18Grabber, 3, ppc_illegalop);
    std::fill_n(Opcode19Grabber, 2048, ppc_illegalop);
    std::fill_n(Opcode31Grabber, 2048, ppc_illegalop);
    std::fill_n(Opcode59Grabber, 64, ppc_illegalop);
    std::fill_n(Opcode63Grabber, 2048, ppc_illegalop);

    #include "ppcopcodes.include"
    #include "poweropcodes.include"
    #include "ppcfpopcodes.include"
}

void ppc_cpu_init(MemCtrlBase* mem_ctrl, uint32_t cpu_version, bool include_601, uint64_t tb_freq)
{
    mem_ctrl_instance = mem_ctrl;

    std::memset(&ppc_state, 0, sizeof(ppc_state));
    set_host_rounding_mode(0);

    ppc_state.spr[SPR::PVR] = cpu_version;
    is_601 = (cpu_version >> 16) == 1;

    initialize_ppc_opcode_tables(include_601);

    // initialize emulator timers
    TimerManager::get_instance()->set_time_now_cb(&get_virt_time_ns);
    TimerManager::get_instance()->set_notify_changes_cb(&force_cycle_counter_reload);

    // initialize time base facility
#ifdef __APPLE__
    mach_timebase_info(&timebase_info);
#endif
    g_nanoseconds_base = cpu_now_ns();
    g_icycles = 0;
    //icnt_factor      = 6;
    icnt_factor = 4;
    tbr_wr_timestamp = 0;
    rtc_timestamp = 0;
    tbr_wr_value = 0;
    tbr_freq_ghz = (tb_freq << 32) / NS_PER_SEC;
    tbr_period_ns = ((uint64_t)NS_PER_SEC << 32) / tb_freq;

    exec_flags = 0;
    exec_timer = false;

    timebase_counter = 0;
    dec_wr_value = 0;


    if (is_601) {
        /* MPC601 sets MSR[ME] bit during hard reset / Power-On */
        ppc_state.msr = (MSR::ME + MSR::IP);
    } else {
        ppc_state.msr             = MSR::IP;
        ppc_state.spr[SPR::DEC_S] = 0xFFFFFFFFUL;
    }

    ppc_mmu_init();

    /* redirect code execution to reset vector */
    ppc_state.pc = 0xFFF00100;

#ifdef CPU_PROFILING
    gProfilerObj->register_profile("PPC_CPU",
        std::unique_ptr<BaseProfile>(new CPUProfile()));
#endif
}

void print_fprs() {
    for (int i = 0; i < 32; i++) {
        cout << "FPR " << dec << i << " : " << ppc_state.fpr[i].dbl64_r << endl;
    }
}

static map<string, int> SPRName2Num = {
    {"XER",    SPR::XER},       {"LR",     SPR::LR},    {"CTR",    SPR::CTR},
    {"DEC",    SPR::DEC_S},     {"PVR",    SPR::PVR},   {"SPRG0",  SPR::SPRG0},
    {"SPRG1",  SPR::SPRG1},     {"SPRG2",  SPR::SPRG2}, {"SPRG3",  SPR::SPRG3},
    {"SRR0",   SPR::SRR0},      {"SRR1",   SPR::SRR1},  {"IBAT0U", 528},
    {"IBAT0L", 529},            {"IBAT1U", 530},        {"IBAT1L", 531},
    {"IBAT2U", 532},            {"IBAT2L", 533},        {"IBAT3U", 534},
    {"IBAT3L", 535},            {"DBAT0U", 536},        {"DBAT0L", 537},
    {"DBAT1U", 538},            {"DBAT1L", 539},        {"DBAT2U", 540},
    {"DBAT2L", 541},            {"DBAT3U", 542},        {"DBAT3L", 543},
    {"HID0",   SPR::HID0},      {"HID1",   SPR::HID1},  {"IABR",   1010},
    {"DABR",   1013},           {"L2CR",   1017},       {"ICTC",   1019},
    {"THRM1",  1020},           {"THRM2",  1021},       {"THRM3",  1022},
    {"PIR",    1023},           {"TBL",    SPR::TBL_S}, {"TBU",    SPR::TBU_S},
    {"SDR1",   SPR::SDR1},      {"MQ",     SPR::MQ},    {"RTCU",   SPR::RTCU_S},
    {"RTCL",   SPR::RTCL_S},    {"DSISR",  SPR::DSISR}, {"DAR",    SPR::DAR},
    {"MMCR0",  SPR::MMCR0},     {"PMC1",   SPR::PMC1},  {"PMC2",   SPR::PMC2},
    {"SDA",    SPR::SDA},       {"SIA",    SPR::SIA},   {"MMCR1",  SPR::MMCR1}
};

static uint64_t reg_op(string& reg_name, uint64_t val, bool is_write) {
    string reg_name_u, reg_num_str;
    unsigned reg_num;
    map<string, int>::iterator spr;

    if (reg_name.length() < 2)
        goto bail_out;

    reg_name_u = reg_name;

    /* convert reg_name string to uppercase */
    std::for_each(reg_name_u.begin(), reg_name_u.end(), [](char& c) {
        c = ::toupper(c);
    });

    try {
        if (reg_name_u == "PC") {
            if (is_write)
                ppc_state.pc = (uint32_t)val;
            return ppc_state.pc;
        }
        if (reg_name_u == "MSR") {
            if (is_write)
                ppc_state.msr = (uint32_t)val;
            return ppc_state.msr;
        }
        if (reg_name_u == "CR") {
            if (is_write)
                ppc_state.cr = (uint32_t)val;
            return ppc_state.cr;
        }
        if (reg_name_u == "FPSCR") {
            if (is_write)
                ppc_state.fpscr = (uint32_t)val;
            return ppc_state.fpscr;
        }
    } catch (...) {
    }

    try {
        if (reg_name_u.substr(0, 1) == "R") {
            reg_num_str = reg_name_u.substr(1);
            reg_num     = (unsigned)stoul(reg_num_str, NULL, 0);
            if (reg_num < 32) {
                if (is_write)
                    ppc_state.gpr[reg_num] = (uint32_t)val;
                return ppc_state.gpr[reg_num];
            }
        }
    } catch (...) {
    }

    try {
        if (reg_name_u.substr(0, 1) == "F") {
            reg_num_str = reg_name_u.substr(1);
            reg_num     = (unsigned)stoul(reg_num_str, NULL, 0);
            if (reg_num < 32) {
                if (is_write)
                    ppc_state.fpr[reg_num].int64_r = val;
                return ppc_state.fpr[reg_num].int64_r;
            }
        }
    } catch (...) {
    }

    try {
        if (reg_name_u.substr(0, 3) == "SPR") {
            reg_num_str = reg_name_u.substr(3);
            reg_num     = (unsigned)stoul(reg_num_str, NULL, 0);
            if (reg_num < 1024) {
                switch (reg_num) {
                case SPR::DEC_U  : reg_num = SPR::DEC_S  ; break;
                case SPR::RTCL_U : reg_num = SPR::RTCL_S ; break;
                case SPR::RTCU_U : reg_num = SPR::RTCU_S ; break;
                case SPR::TBL_U  : reg_num = SPR::TBL_S  ; break;
                case SPR::TBU_U  : reg_num = SPR::TBU_S  ; break;
                }
                if (is_write)
                    ppc_state.spr[reg_num] = (uint32_t)val;
                return ppc_state.spr[reg_num];
            }
        }
    } catch (...) {
    }

    try {
        if (reg_name_u.substr(0, 2) == "SR") {
            reg_num_str = reg_name_u.substr(2);
            reg_num     = (unsigned)stoul(reg_num_str, NULL, 0);
            if (reg_num < 16) {
                if (is_write)
                    ppc_state.sr[reg_num] = (uint32_t)val;
                return ppc_state.sr[reg_num];
            }
        }
    } catch (...) {
    }

    try {
        spr = SPRName2Num.find(reg_name_u);
        if (spr != SPRName2Num.end()) {
            if (is_write)
                ppc_state.spr[spr->second] = (uint32_t)val;
            return ppc_state.spr[spr->second];
        }
    } catch (...) {
    }

bail_out:
    throw std::invalid_argument(string("Unknown register ") + reg_name);
}

uint64_t get_reg(string reg_name) {
    return reg_op(reg_name, 0, false);
}

void set_reg(string reg_name, uint64_t val) {
    reg_op(reg_name, val, true);
}
